Pro mike_flatfixer,flatlist,wavearr,slitarr,orderlist,combflat,combflatvar,$
                   blue=blue,red=red,milky=milky,gain=gain,rdnoise=rdnoise,$
                   illum=illum,bkspace=bkspace
                   
;+-------------------------------------------------------------------
;
; MIKE_FLATFIXER    2/2007
;
; Median-combine and normalize MIKE flats.
;
; Input flats are median combined, the combined flat is normalized,
; and the formal variance in the normalized flat is computed assuming
; Poisson statistics.
;
; Assumes last 128 rows and columns are overscan (should be OK for
; any binning).
;
; Must set either /blue or /red keyword.
;
; Break point spacing for b-spline is set depending on the type of
; flat.  Assumptions are as follows.  The break point spacing is
; chosen to be small enough to capture the intrinsic spectral shape of
; the source, but large enough so as not to fit out intrinsic flat
; field features.  There is some doubt whether flat field features
; that appear towards the bottom of the CCD in non-milky blue flats
; are real.  These do not appear in the blue milky flats, but they may
; be getting fit out.
;       CCD        Milky?       Source           Spacing
;       ------------------------------------------------
;       blue       no           Int Flat Lamp    100
;       red        no           Int Flat Lamp    100
;       blue       yes          Star              30
;       red        yes          Int Flat Lamp    100
;
; Errors in the flat field may be ~1 to a few % on the scales of the
; break point spacing, but should be near Poisson limit for pixel-to-
; pixel variations.
;
; INPUTS
;       flatlist  - String, name of single-column ASCII file containing
;                   the FITS filenames of flats to be combined.
;       wavearr   - 2D array or fits filename, where the pixel values
;                   are uncalibrated wavelengths (in pixels).
;       slitarr   - 2D array or fits filename, where the pixel values
;                   are position along the slit.
;       orderlist - Name of ASCII file containing the order
;                   information [CCD, order #, center, left edge,
;                   right edge], as output by mike_slitgrid.pro.
;
; KEYWORDS
;       blue      - Blue CCD.
;       red       - Red CCD.
;       milky     - If set, assumes these are mily flats.  The entire
;                   illuminated region of the CCD is fitted, including
;                   regions between orders.  A wide break point spacing
;                   and a high-order slit-illumination function are used.
;                   Default is to assume non-diffused flats.
;       gain      - Scalar, gain in e-/ADU.  Defaults are set separately 
;                   for blue and red CCDs
;       rdnoise   - Scalar, read noise, in e-.  Defaults are set separately 
;                   for blue and red CCDs
;       illum     - Scalar, polymonial order+1 used to fit the slit
;                   illumination functin (1=constant, 2=linear, 3=
;                   quadrati, etc.).  Default is 2 for non-diffused flats
;                   and 5 for milky flats.
;       bkspace   - Spacing of bspline break points, in pixels (or whatever
;                   units are used for the input wavearr).  Defaults are
;                   separately for blue and red CCDs, and for milky and
;                   non-diffused flats.
;                   Smaller spacing may be needed for exposures
;                   that contain emission lines.  In general, however, a
;                   large spacing should be used so that fringe features
;                   are not divided out.
;
;  OUTPUTS
;       combflat    - Median-combined, normalized flat field.  May
;                     either be a variable or a fits filename to which
;                     array is to be written.
;       combflatvar - Formal variance in the combined, normlized flat.
;                     May either be a variable or a fits filename to
;                     which array is to be written.
;
; HISTORY
;       Adapted from LDSS3_FLATFIXER 2/18/2007 GDB
;---------------------------------------------------------------------

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE:  '
   print,' mike_flatfixer,flatlist,wavearr,slitarr,orderlist,combflat,'
   print,'                combflatvar,/blue,/red,/milky,gain=gain,'
   print,'                rdnoise=rdnoise,illum=illum,bkspace=bkspace'
   return
endif

;;; Determine whether this is the blue or red CCD, and set
;;; CCD-specific parameters.
if keyword_set(blue) then begin
   blue = 1
   red  = 0
   gn   = 0.47    ; e-/ADU
   rdn  = 2.00    ; e-
endif else begin
   if keyword_set(red) then begin
      blue = 0
      red  = 1
      gn   = 1.00
      rdn  = 3.50
   endif else begin
      print,'MIKE_FLATFIXER: Must set either the /blue or /red keyword.'
      return
   endelse
endelse

;;; Read in order information
order_templ =  {   $
          version: 1.0, $
          datastart: 0, $
          delimiter: ' ', $        ; Fields separated by white space
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 5, $
          fieldtypes: [7,4,4,4,4], $
          fieldnames: ['ccd','ordernum','center','left','right'], $
          fieldlocations: [2,11,18,30,42], $
          fieldgroups: [0,1,2,3,4] $
        }
orderinfo       = read_ascii(orderlist,template=order_templ)
tmp_ordernum    = orderinfo.ordernum
tmp_ordercenter = orderinfo.center
tmp_orderedge1  = orderinfo.left
tmp_orderedge2  = orderinfo.right
; Sort orders in order of increasing slit position
order_order = sort(tmp_ordercenter)
ordernum    = tmp_ordernum(order_order)
ordercenter = tmp_ordercenter(order_order)
orderedge1  = tmp_orderedge1(order_order)
orderedge2  = tmp_orderedge2(order_order)
n_orders    = n_elements(ordercenter)

;;; Read in flat filenames
flat_templ = {   $
       version: 1.0, $
       datastart: 0L, $
       delimiter: ' ', $
       missingvalue: -99.0, $
       commentsymbol: '#', $
       fieldcount: 1L, $
       fieldtypes: [7L], $
       fieldnames: ['name'], $
       fieldlocations: [6L], $
       fieldgroups: [0L] $
     }
list     = read_ascii(flatlist,template=flat_templ)
flatfile = list.name
nflat    = n_elements(flatfile)

;;; Read in wave and slit grids.

if (size(wavearr,/tname) eq 'STRING') then begin
   checkexists = findfile(wavearr,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Wavelength file not found.'
      return
   endif else begin
      wavegrid = readfits(wavearr)
   endelse
endif else wavegrid = wavearr

if (size(slitarr,/tname) eq 'STRING') then begin
   checkexists = findfile(slitarr,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Slit position file not found.'
      return
   endif else begin
      slitgrid = readfits(slitarr)
   endelse
endif else slitgrid = slitarr

;;;
;;; Median combine flats
;;;

; Arrays to store processed flats.
sz         = size(wavegrid)
xsz        = sz(1)
ysz        = sz(2)
bigflat    = fltarr(xsz,ysz,nflat)
texp       = fltarr(nflat)         ; Exposure times

; Read in flats
for k=0,nflat-1 do begin
   tmp_flatfile = flatfile(k)
   checkexists  = findfile(tmp_flatfile,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Flat file '+tmp_flatfile+' not found.'
      return
   endif else begin
      tmp_flat       = readfits(tmp_flatfile,flat_hdr)
      ; Subtract bias
      bigflat(*,*,k) = mike_subbias(tmp_flat)
      texp(k)        = sxpar(flat_hdr,'EXPTIME')
   endelse
endfor

; Check that all flats have similar exposure times
texp_fracdiff = abs(texp - mean(texp))/mean(texp)
if (max(texp_fracdiff) gt 0.01) then begin
   print,'MIKE_FLATFIXER: Different exposure times detected.'
   return
endif

; Median combine
if (nflat gt 1) then begin
   medflat = median(bigflat,dim=3,/even)
endif else begin
   medflat = bigflat
endelse

;;;
;;; Normalize
;;;

; User-specified gain and read noise (optional)
if keyword_set(gain) then gn = gain         ; e-/ADU
if keyword_set(rdnoise) then rdn = rdnoise  ; e-
   
; Set break point spacing for b-spline fit.
if keyword_set(bkspace) then bk_space = bkspace else begin
   if keyword_set(milky) then begin
      if blue then bk_space = 30 else bk_space = 100.
   endif else bk_space = 100.
endelse

; Polynomial order for slit illumination function
if keyword_set(illum) then begin
   illum_order = illum 
endif else begin
   if keyword_set(milky) then illum_order = 5 else illum_order = 2
endelse

; Arrays of x and y-values
xarr = lonarr(xsz,ysz)
for i=0,xsz-1 do xarr(i,*) = i
yarr = lonarr(xsz,ysz)
for i=0,ysz-1 do yarr(*,i) = i
illum_xsz = xsz - 128
illum_ysz = ysz - 128

; Arrays of bad pixels/rows/columns
badarr = lonarr(xsz,ysz)
if blue then begin 
   badarr(630:632,899:ysz-1) = 1
   badarr(912,1412:ysz-1)    = 1
   badarr(959,1468:ysz-1)    = 1
endif

; Output arrays
normflat     =  1 + fltarr(xsz,ysz)
normflat_var = -1 + fltarr(xsz,ysz)

; Array of edges, useful for non-diffused flats 
; Note: Running SOBEL will also identify skylines as edges.
xmask      = [[-1,0,1],[-2,0,2],[-1,0,1]]
edge_array = convol(medflat,xmask)/float(medflat)

;;; For each order, identify illuminated region and fit.

for i=0,n_orders-1 do begin

   print,' '
   print,'Order '+strtrim(ordernum(i),2)

   ; Identify illumated region
   if (not(keyword_set(milky)) or n_orders eq 1) then begin
      ; For non-diffused flats, include only regions within the trace
      slit_lo = orderedge1(i)
      slit_hi = orderedge2(i)
   endif else begin
      ; For milky flats, include regions outside the trace
      case i of
         0:    begin
                  slit_hi = 0.5*(orderedge2(i)+orderedge1(i+1))
                  slit_lo = slit_hi - (ordercenter(i+1)-ordercenter(i))
               end
         n_orders-1 : begin
                  slit_lo = 0.5*(orderedge2(i-1)+orderedge1(i))
                  slit_hi = slit_lo + (ordercenter(i)-ordercenter(i-1))
               end
         else: begin
                  slit_lo = 0.5*(orderedge2(i-1)+orderedge1(i))
                  slit_hi = 0.5*(orderedge2(i)+orderedge1(i+1))
               endelse
      endcase
   endelse
   illum_ls = where(slitgrid ge slit_lo and slitgrid lt slit_hi and $
                    xarr le illum_xsz-1 and yarr le illum_ysz-1,n_illum)
   
   ; Identify region to be fit.  If this is a non-diffused flat, then 
   ; avoid edge regions.
   if keyword_set(milky) then begin
      fit_ls = where(slitgrid ge slit_lo and slitgrid lt slit_hi and $
                       xarr le illum_xsz-1 and yarr le illum_ysz-1 and $
                       badarr ne 1,n_fit)
   endif else begin
      fit_ls = where(slitgrid ge slit_lo and slitgrid lt slit_hi and $
                     xarr le illum_xsz-1 and yarr le illum_ysz-1 and $
                     abs(edge_array) lt 2 and badarr ne 1,n_fit)
   endelse
   fitslit = slitgrid(fit_ls)
   fitwave = wavegrid(fit_ls)
   fitflat = medflat(fit_ls)
   ; Assume Poisson noise for initial variance estimate
   if (nflat gt 1) then med_factor = !pi/2. else med_factor = 1
   fitflat_var = med_factor*(fitflat/gn + (rdn/gn)^2)/nflat
   fitflat_invvar = 1./fitflat_var
   
   ; Identify potential bad pixels before fitting.
   print,'Pre-filtering bad pixels...'
   maxwave = max(fitwave)
   minwave = min(fitwave)
   first_filter = 1
   filter_sz = 2*bk_space
   n_filters = ceil((maxwave-minwave)/filter_sz)
   ; Default is not to filter any points
   filter_ls = -1
   ; Speed things up by dividing arrays into regions that contain
   ; several filter bins.
   filters_per_region = 1.*round(sqrt(n_filters))
   num_regions        = ceil(1.*n_filters / filters_per_region)
   for r=0,num_regions-1 do begin
      j_lo = r*filters_per_region
      j_hi = ((r+1)*filters_per_region - 1) < (n_filters - 1)
      region_wavelo = minwave + j_lo*filter_sz
      region_wavehi = minwave + (j_hi+1)*filter_sz
      region = where(fitwave ge region_wavelo and $
                     fitwave lt region_wavehi,n_region)
      if (n_region gt 0) then begin
         region_wave = fitwave(region)
         region_flat = fitflat(region)
         for j=j_lo,j_hi do begin
            wavelo = minwave + j*filter_sz
            wavehi = minwave + (j+1)*filter_sz
            bin_ls = where(region_wave ge wavelo and region_wave le wavehi,$
                           n_bin)
            if (n_bin gt 1) then begin
               tmp_flat   = region_flat(bin_ls) 
               tmp_median = median(tmp_flat)
               tmp_sdev   = robust_sigma(tmp_flat)
               tmp_filter_ls = where(abs(tmp_flat-tmp_median) ge 5*tmp_sdev or $
                                     tmp_flat lt tmp_sdev,n_tmp_filter)
               if (n_tmp_filter gt 0) then begin
                  if first_filter then begin
                     filter_ls    = region(bin_ls(tmp_filter_ls))
                     first_filter = 0 
                  endif else begin
                     filter_ls = [filter_ls,region(bin_ls(tmp_filter_ls))]
                  endelse
               endif
            endif
         endfor
      endif
   endfor
   if (total(filter_ls) ne -1) then fitflat_invvar(filter_ls) = 0
   
   ; Iterate b-spline fit
   niter   = 7
   good_ls = lindgen(n_fit)
   ; Start with a uniform illumination function
   illum_model = 1. + fltarr(n_fit)
   for k=1,niter do begin   
      if (k eq 1) then begin
         ; Do a preliminary fit by taking the median in each bin.
         ; Might have to increase the bin size for this step.
         h = histogram(fitwave,bin=bk_space,locations=wbin,rev=ri)
         nbin = n_elements(h)
         fitflat_model = 0*fitflat
         for j=0,nbin-1 do begin
            bin_ls = ri(ri(j):ri(j+1)-1)  
            if (h(i) ne 0) then begin
               tmp_flat   = fitflat(bin_ls)
               tmp_invvar = fitflat_invvar(bin_ls)
               nonzero_ls = where(tmp_invvar gt 0,n_nonzero)
               if (n_nonzero ne 0) then $
                  fitflat_model(bin_ls) = median(tmp_flat(nonzero_ls)) else $
                  fitflat_model(bin_ls) = median(tmp_flat)
            endif else begin
               print,'No pixels for intial estimate'
               stop
            endelse
         endfor
      endif else begin
         ; Fit counts with illumination function removed
         sset = bspline_iterfit(fitwave(good_ls),$
                                (fitflat/illum_model)(good_ls),$
                                invvar=fitflat_invvar(good_ls),$
                                bkspace=bk_space,$
                                yfit=fitflat_model_good,$
                                upper=5,lower=5)
         ; Evaluate fit over entire fit region
         fitflat_model_no_illum = bspline_valu(fitwave,sset)      
         ; Compute new illumination function
         if (illum_order ge 2) then begin
            pfit = robust_poly_fit(fitslit,fitflat/fitflat_model_no_illum,$
                                   illum_order-1,illum_model)
            ; Apply new illumination function
            fitflat_model = fitflat_model_no_illum * illum_model
         endif else begin
            ; No illumination function correction
            fitflat_model = fitflat_model_no_illum
         endelse
      endelse
      ; Compute variance.  Using this method is useful when
      ; the scatter about the fit is larger than the formal
      ; variance for a given pixel would suggest, as is the
      ; case when there is fringing.
      h  = histogram(fitwave,bin=50,rev=ri)
      nbin = n_elements(h)
      fitflat_var = 1e6 + 0.*fitflat
      fitdiff     = fitflat - fitflat_model
      for j=0,nbin-1 do begin
         bin_ls = ri(ri(j):ri(j+1)-1)
         if (h(i) ne 0) then $
            fitflat_var(bin_ls) = (robust_sigma(fitdiff(bin_ls)))^2
      endfor
      zerovar_ls = where(fitflat_var eq 0,n_zerovar,compl=nonzerovar_ls)
      if (n_zerovar ne 0) then $
         fitflat_var(zerovar_ls) = median(fitflat_var(nonzerovar_ls))
      fitflat_invvar = 1./fitflat_var
      ; Reject outliers
      good_ls = where(abs(fitflat-fitflat_model) le 5*sqrt(fitflat_var))
   endfor
stop
   ; Evaluate bspline fit over illuminated_region and divide out
   medflat_model_no_illum = fltarr(xsz,ysz)
   medflat_model_no_illum(illum_ls) = bspline_valu(wavegrid(illum_ls),sset)
   if (illum_order ge 2) then begin
      medflat_illum = fltarr(xsz,ysz)
      for k=0,illum_order-1 do medflat_illum(illum_ls) = $
                 medflat_illum(illum_ls) + pfit(k)*(slitgrid(illum_ls)^k)
      medflat_model = medflat_model_no_illum * medflat_illum
   endif else begin
      medflat_model = medflat_model_no_illum
   endelse
   normflat(illum_ls) = medflat(illum_ls) / medflat_model(illum_ls)

   ; Estimate variance in the normalized flat.  Note: this estimate will
   ; be wrong in the under-illuminated corners!  Non-illuminated regions 
   ; are flagged with -1.
   normflat_var(illum_ls) = $
         (med_factor*(medflat_model(illum_ls)/gn + (rdn/gn)^2)/nflat) / $
         (double(medflat_model(illum_ls))^2)

endfor

;;;
;;; Set output arrays or write fits files
;;;

hnormflat = flat_hdr
sxaddhist,'-------------------------------------------------------',hnormflat
sxaddhist,'Flatfield processing done using MIKE_FLATFIXER',hnormflat
sxaddhist,'   Flatfield frames combined:',hnormflat
for i=0,nflat-1 do $
   sxaddhist,'      '+flatfile(i),hnormflat
sxaddhist,'-------------------------------------------------------',hnormflat

sxaddpar,hnormflat,'BZERO',0.0

if (size(combflat,/tname) eq 'STRING') then begin
   print,'Writing '+combflat
   writefits,combflat,normflat,hnormflat
endif else combflat = normflat

if (size(combflatvar,/tname) eq 'STRING') then begin
   print,'Writing '+combflatvar
   writefits,combflatvar,normflat_var,hnormflat
endif else combflatvar = normflat_var

print,'Done.'

return
end
