Pro mike_reduce,object,wavefile,slitfile,orderlist,arc=arc,$
                flatfield=flatfield,flatvar=flatvar,output=output,$
                findrange=findrange,order=order,fit=fit,fiducial=fiducial,$
                object_xy=object_xy,fid_mask=fid_mask,scattered=scattered,$
                _extra=ex

;+-------------------------------------------------------------------
;
; MIKE_REDUCE     2/2007
;
; Reduce a MIKE frame using mutliple calls to longslit_reduce.
;
; INPUTS
;         object      - Object FITS filename
;         wavefile    - FITS filename of calibrated wavelength map
;                       (see mike_wavegrid.pro and
;                       mike_wavecalib.pro).  These should be AIR
;                       wavelengths.  Output wavelengths will be
;                       vacuum heliocentric.
;                       *** Currently no corrections made using skylines. ***
;         slitfile    - FITS filename of slit position map (see
;                       mike_slitgrid.pro)
;         orderlist   - Name of 5-column ASCII file containing, for
;                       each order, the the CCD name (blue or red),
;                       the order number (which is arbitrary, but
;                       should be consecutive integers), and should
;                       not be 0), the minimum slit position covered
;                       by the order, and the maximum slit position
;                       covered by the order.  Format is: [CCD, order
;                       #, center, left edge, right edge].  Output by
;                       mike_slitgrid.pro.
;
; KEYWORDS
;         arc         - FITS filename of arc exposure.  If set,
;                       wavelength solution for this arc will be
;                       performed using the input wavefile as a
;                       reference.  Default is to assume the wavelengths
;                       in the input wavefile are accurate for this
;                       exposure.  Wavelength solution for this arc
;                       should be within a resolution element of the
;                       input wavefile.
;         flatfield   - FITS filename of flat field array
;         flatvar     - FITS filename of flat field variance array
;         output      - Root name for output files.  Default is
;                       'mike'.
;         findrange   - Two-elements array giving the min and max y-values
;                       over which to locate the object in the fiducial
;                       order.  DISABLED.
;         order       - 1D array specifying the orders to be reduced.
;                       Numbers correspond to the entry number in the
;                       input orderlist (1 being the first order listed
;                       in the file).  Must include the fiducial order.
;         fit         - 1D array specifying the orders in which the object
;                       should be indepently located and traced, and its
;                       FWHM measured.  These orders should have significant
;                       flux over most of the length of the array.  Other
;                       orders will have their object position, trace, and
;                       FWHM extrapolated from the fitted orders, which
;                       is useful for orders with little or no flux.  Default
;                       is to independently fit all orders.  Must include
;                       the fiducial order.
;         fiducial    - Scalar, order number to be used as the
;                       fiducial order (the order number is determined
;                       by the order number entry in the input
;                       orderlist).  This order will be reduced first
;                       and willl be used to measure the initial
;                       profile FWHM (for Gaussian profiles).  The
;                       default is to use one of the middle orders.
;                       However, it is recommended to use the order
;                       that contains the most flux.
;         object_xy   - Two-element array giving the [x,y] pixel
;                       coordinates of a point at the center of the
;                       object trace in the the fiducial order.
;                       Useful when more than one object appears on
;                       the slit.  If set, the program uses this as an
;                       initial estimate but refines the position
;                       using a Gaussian fit.  Default is to automatically
;                       search for the brightest object in the slit.
;         fid_mask    - Three-element array specifying a region on the
;                       fiducial order to mask.  Useful when multiple
;                       objects appear on the slit.  The array
;                       elements describe a horizontal region with
;                       [xmin,xmax,y].  The regions in other orders
;                       corresponding to these slit positions will
;                       also be masked.  Note that it is not necessary
;                       to set this keyword for single-object
;                       exposures, provided the object can be
;                       automatically located.
;         scattered   - Subtract scattered light by fitting a 2D polynomial
;                       to the inter-order counts.
;         _extra      - Additional keywords accepted by
;                       longslit_reduce, e.g., sprofile, bright, maxiter,
;                       noshow, and noskysub.  Note that some keywords
;                       may not work due to their use in the program
;                       body below.
;
; OUTPUTS
;         Outputs written to output_sub.fits, output_spec.fits, etc.
;
; Side effects: If there is a journal file currently open, the program
;               will close it and open <output>.log.
;
; HISTORY
;         Adapted from HIRES_REDUCE 2/26/2007 GDB
;         ARC keyword added 5/14/2007 GDB
;         Fixed method by which higher order profile and trace terms
;            are mapped to the current exposure.  4/17/2009 GDB
;         Convert wavelengths to vacuum 4/22/2009 GDB
;         Adjusted break point spacing for sky b-spline fit,
;            OBJECT_XY keyword added 1/3/2013 GDB
;--------------------------------------------------------------------------

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE:'
   print,' mike_reduce,object,wavefile,slitfile,orderlist,arc=arc,'
   print,'        flatfield=flatfield,flatvar=flatvar,output=output,'
   print,'        findrange=findrange,order=order,fit=fit,fiducial=fiducial,'
   print,'        object_xy=object_xy,fid_mask=fid_mask,scattered=scattered,'
   print,'        _extra=ex'
   return
endif

; Root name for output
if keyword_set(output) then root = output else root = 'mike'

; Start log.  If there is a journal already running, close it and start a
; new one.
if (!journal ne 0) then journal
journal,root+'.log'

print,''
print,'Reducing '+object
print,''

; Read in FITS files
obj  = readfits(object,hobj)
wave = readfits(wavefile,hwave)
slit = readfits(slitfile,hslit)

; Bias-subtract
obj = mike_subbias(obj)

; Get size of individual CCD arrays
sz  = size(obj)
xsz = sz(1)
ysz = sz(2)

; Read in flat files.  Assume bias has already been subtracted.
if keyword_set(flatfield) then flat = readfits(flatfield,hflat) $
   else flat = 0
if keyword_set(flatvar) then flat_var = readfits(flatvar,hflatvar) $
   else flat_var = 0

; Read in order information
order_templ =  {   $
          version: 1.0, $
          datastart: 0, $
          delimiter: ' ', $        ; Fields separated by white space
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 5, $
          fieldtypes: [7,4,4,4,4], $
          fieldnames: ['ccd','ordernum','center','left','right'], $
          fieldlocations: [2,11,18,30,42], $
          fieldgroups: [0,1,2,3,4] $
        }
orderinfo       = read_ascii(orderlist,template=order_templ)
; Make sure all order numbers are unique
if (n_elements(orderinfo.ordernum) ne $
    n_elements(uniq(orderinfo.ordernum))) then begin
   print,'Error - All order numbers must be unique.'
   return
endif
; Sort orders in order of increasing slit position
sort_ls     = sort(orderinfo.center)
ccdname     = orderinfo.ccd(sort_ls)
ordernum    = orderinfo.ordernum(sort_ls)
ordercenter = orderinfo.center(sort_ls)
orderedge1  = orderinfo.left(sort_ls)
orderedge2  = orderinfo.right(sort_ls)
n_orders    = n_elements(ordercenter)

; Optional - Specify the orders to reduce.
if keyword_set(order) then begin
   n_reduce    = n_elements(order)
   reduce_ls   = intarr(n_reduce)
   for i=0,n_reduce-1 do reduce_ls(i) = where(ordernum eq order(i))
   ordercenter = ordercenter(reduce_ls)
   ccdname     = ccdname(reduce_ls)
   ordernum    = ordernum(reduce_ls)
   orderedge1  = orderedge1(reduce_ls)
   orderedge2  = orderedge2(reduce_ls)
   n_orders    = n_elements(ordercenter)
endif

; Optional - Specify orders the orders for which to independently
; locate and trace the object, and measure its FWHM.  Default is
; to independently fit all orders.
do_independent_fit = lonarr(n_orders)
if keyword_set(fit) then begin
   for i=0,n_elements(fit)-1 do begin
      j = where(ordernum eq fit(i))
      do_independent_fit(j) = 1
   endfor
endif else begin
   do_independent_fit(*) = 1
endelse

; Keep track of orders for which independent tracing and fitting
; has been done.
independent_fit_made = lonarr(n_orders)

; Optional - Specify the order to be reduced first.  Default is
; to start with the lowest-numbered order.
if keyword_set(fiducial) then begin
   fid_index = where(ordernum eq fiducial,n_fid)
   fid_index = fid_index(0)
   if (n_fid eq 0) then begin
      print,'Error - Fiducial order must be included in the orders to reduce.'
      return
   endif
endif else fid_index = floor(n_orders/2.)

; Optional - Create wavelength solution from arc exposure
; using the input wavefile as a referece.
if keyword_set(arc) then begin
   print,'Computing wavelength solution from arc...'
   mike_wavecalib,arc,orderlist,slitfile,wavefile,'temp_mike_wavegrid.fits',$
                  /tweak,/auto
   wave = readfits('temp_mike_wavegrid.fits',hwave)
   spawn,'"rm" temp_mike_wavegrid.fits'
endif

; Convert wavelengths to vacuum
print,'Converting wavelengths to vacuum...'
airtovac,wave

; Flag bad columns to be masked by longslit_reduce.
; Non-illuminated regions should have been masked by mike_flatfixer.
case ccdname(0) of
   'blue': begin
              print,'Masking bad columns on blue CCD'
              flat(629:632,898:ysz-1) = 0
              flat(633,898:906)       = 0
              flat(912,1412:ysz-1)    = 0
              flat(959,1468:ysz-1)    = 0
           end
   'red' : ; No masking
   else  : print,'WARNING: CCD name not recognized.  No masking done.'
endcase

; Optional - Specify a region along the fiducial order to be masked
; when performing the sky subtraction.  The corresponding regions
; in other orders will also be masked.
if keyword_set(fid_mask) then begin
   xmin     = fid_mask(0)
   xmax     = fid_mask(1)
   ymin     = fid_mask(2)
   ymax     = fid_mask(2)
   ; Min/max slit positins on fiducial order to mask
   slit_cut = slit(xmin:xmax,ymin:ymax)
   fid_mask_slit_min = min(slit_cut)
   fid_mask_slit_max = max(slit_cut)
   ; Min/max offset from order center
   mask_offset_lo = fid_mask_slit_min - ordercenter(fid_index)
   mask_offset_hi = fid_mask_slit_max - ordercenter(fid_index)
   apply_mask     = 1
endif else begin
   apply_mask     = 0
endelse

; Optional - Set y-range over which to locate object.
if keyword_set(findrange) then nominal_find_range=findrange else begin
   case ccdname(0) of
      'blue': nominal_find_range=[600,800]
      'red' : nominal_find_range=[1100,1300]
      else  : begin
                 print,'Error - CCD name not recognized'
                 return
              end
   endcase
endelse

; Optional - Subtract scattered light
if keyword_set(scattered) then begin
   print,'Subtracting scattered light...'
   ; Select regions in between orders
   maskval    = -100
   interorder = maskval + 0.*obj
   for i=1,n_orders-2 do begin
      inter_ls = where(slit ge orderedge2(i)+1   and $
                       slit le orderedge1(i+1)-1,n_inter)
      if (n_inter ne 0) then interorder(inter_ls) = obj(inter_ls)
   endfor
   ;; Mask cosmic rays
   ;cr_ls = where(interorder ge 200,n_cr)  ; Arbitrary count threshold
   ;if (n_cr ne 0) then begin
   ;   cr_xy = array_indices(interorder,cr_ls)
   ;   for i=0,n_cr-1 do begin
   ;      xlo = (cr_xy(0,i)-1) > 0
   ;      xhi = (cr_xy(0,i)+1) < (xsz-1)
   ;      ylo = (cr_xy(1,i)-1) > 0
   ;      yhi = (cr_xy(1,i)+1) < (ysz-1)
   ;      interorder(xlo:xhi,ylo:yhi) = maskval
   ;   endfor
   ;endif
   ; Mask non-illuminated regions
   xarr = fltarr(xsz,ysz)
   yarr = fltarr(xsz,ysz)
   for i=0,xsz-1 do xarr(i,*) = i
   for i=0,ysz-1 do yarr(*,i) = i
   nonillum_ls = where(xarr ge 1024 or yarr ge 2048,n_nonillum)  ; 2x2 binning
   if (n_nonillum ne 0) then interorder(nonillum_ls) = maskval
   ; Fit 2D polynomial
   fit_ls = where(interorder ne maskval,n_fit)
   if (n_fit eq 0) then begin
      print,'MIKE_REDUCE: No good pixels for fitting scattered light.'
      return
   endif
   interorder_data = fltarr(3,n_fit)
   interorder_data(0:1,*) = array_indices(interorder,fit_ls)
   interorder_data(2,*)   = interorder(fit_ls)
   fit_order = 4
   temp_fit  = sfit(interorder_data,fit_order,/irregular,kx=fitterms)
   ; Evaluate polynomial over entire array and subtract
   scattered_light = 0.*obj
   for i=0,fit_order do begin
      for j=0,fit_order do begin
         scattered_light = scattered_light + fitterms(j,i)*(xarr^i)*(yarr^j)
      endfor
   endfor
   if (n_nonillum ne 0) then scattered_light(nonillum_ls) = 0.
   obj = obj - scattered_light
endif

; Initalize large spectrum, error, and wavelength
; arrays in which to store the individual reduced orders.
big_wave = dblarr(10000,n_orders)
big_spec = dblarr(10000,n_orders)
big_err  = dblarr(10000,n_orders)

; Keep track of the median wavelength, Gaussian profile FWHM, slit
; position, mean wavelength shift and trace coefficients (if tracing
; object) for each order.
medwave_list   = dblarr(n_orders)
fwhm_list      = dblarr(n_orders)
position_list  = dblarr(n_orders)
meanshift_list = dblarr(n_orders)
trace_coeffs   = dblarr(5,n_orders)

; Initialize fits to FWHM, position, and linear trace term
fwhm_fit     = dblarr(n_orders)
position_fit = dblarr(n_orders)
slope_fit    = dblarr(n_orders)

; Initialize reduced arrays
nosky    = fltarr(xsz,ysz)
noskyvar = fltarr(xsz,ysz)
wavefit  = dblarr(xsz,ysz)
skymodel = fltarr(xsz,ysz)

; Set direction of increasing wavelength.  Start by stepping through
; orders in the direction of increaseing wavelength.
case ccdname(0) of
   'blue': begin
              red_direction  = -1
              blue_direction =  1
           end
   'red' : begin
              red_direction  =  1
              blue_direction = -1
           end
endcase

; High-order terms for the fwhm, position, and trace slope as a
; function of ordernumber for a reference standard star, for which the
; order numbering may not be the same as the current exposure.  These
; are for slitgrids created from tracing bright standards at low
; airmass.  The standard star used to determine the higher order terms
; was Nov 2006, exposure 0077, which had the center of its order 20
; (lambda_mid = 6606) near x=615 in the red, and the center of its
; order 20 (lambda_mid = 3956) near x=502 in the blue.
ref_ordernum    = findgen(45) - 3
case ccdname(0) of
   'blue': begin
              ref_fwhm_hot     = (  -0.0327275) * ref_ordernum^1 + $
                                 (  0.00490972) * ref_ordernum^2 + $
                                 (-0.000247102) * ref_ordernum^3 + $
                                 ( 5.45696e-06) * ref_ordernum^4
              ref_position_hot = (     20.4990) * ref_ordernum^1 + $
                                 (    0.149468) * ref_ordernum^2 + $
                                 ( 0.000630445) * ref_ordernum^3 + $
                                 ( 8.21349e-06) * ref_ordernum^4
              ref_slope_hot    = (-5.18004e-07) * ref_ordernum^1
           end
   'red' : begin
              ref_fwhm_hot     = (  -0.0666871) * ref_ordernum^1 + $
                                 (  0.00350955) * ref_ordernum^2 + $ 
                                 (-6.39181e-05) * ref_ordernum^3
              ref_position_hot = (     41.5683) * ref_ordernum^1 + $
                                 (   -0.483866) * ref_ordernum^2 + $
                                 (  0.00513571) * ref_ordernum^3 + $ 
                                 (-6.90471e-05) * ref_ordernum^4 + $ 
                                 ( 8.41102e-07) * ref_ordernum^5
              ref_slope_hot    = ( 4.62247e-05) * ref_ordernum^1 + $
                                 (-5.47953e-06) * ref_ordernum^2 + $
                                 ( 2.42483e-07) * ref_ordernum^3 + $ 
                                 (-3.56052e-09) * ref_ordernum^4
           end
endcase
; Match the higher order terms to the current exposure by identifying
; the order in the current exposure for which the middle wavelength is
; the same as that for the reference standard star order 20.
slit_slice = slit(*,1024)
wave_slice = wave(*,1024)
wave_mid   = interpol(wave_slice,slit_slice,ordercenter)
if (ccdname(0) eq 'blue') then ref_wave_mid = 3956. else ref_wave_mid = 6606.
i_match    = where(abs(wave_mid-ref_wave_mid) eq min(abs(wave_mid-ref_wave_mid)))
i_match    = i_match(0)
ordernum_match  = ordernum(i_match)
ordernum_offset = ordernum_match - 20
fwhm_hot        = interpol(ref_fwhm_hot,ref_ordernum,ordernum-ordernum_offset)
position_hot    = interpol(ref_position_hot,ref_ordernum,ordernum-ordernum_offset)
slope_hot       = interpol(ref_slope_hot,ref_ordernum,ordernum-ordernum_offset)

; Loop through orders.  Start at the fiducial order and proceed first to
; redder orders, then to bluer.
index      = fid_index
index_step = red_direction
done       = 0

while not(done) do begin

   print,' '
   print,'Reducing '+object+', order '+strtrim(ordernum(index),2)
   print,' '

   ; Extrapolate fit and trace parameters from orders that have been
   ; independently fit.  If the higher order terms do not work for
   ; other settings or runs, may need to redo higher order fits below.
   ; If there are few orders for which these quantities can be
   ; measured independently, do a constant offset only.  This is
   ; equivalent to doing an offset from the standard star from which
   ; these terms were measured.  In that case, the extrapolated values
   ; may only be reasonable out to ~10 additional orders.  When many
   ; orders are fit independently, fit a linear trend to the reisduals
   ; of the fitted values minus the higher order terms.
   
   ; When fitting, ignore the reddest orders, which seem to be
   ; problematic.  Note: Selection of independent orders for fit
   ; should be the same when outputing *trace.ps file.
   indep_ls     = where(independent_fit_made eq 1 and $
                        medwave_list le 8900,n_indep)
   n_for_linear = 6
   if (n_indep ge n_for_linear) then begin
      fwhm_order     = 1
      position_order = 1
      slope_order    = 1
   endif else begin
      fwhm_order     = 0
      position_order = 0
      slope_order    = 0
   endelse
   if (n_indep ge 1) then begin
      ; Subtract high-order terms
      net_fwhm     = fwhm_list         - fwhm_hot
      net_position = position_list     - position_hot
      net_slope    = trace_coeffs(1,*) - slope_hot
      ; Fit residuals
      if (n_indep ge n_for_linear) then begin
         pfit_fwhm     = robust_poly_fit(ordernum(indep_ls),$
                                         net_fwhm(indep_ls),$
                                         fwhm_order)
         pfit_position = robust_poly_fit(ordernum(indep_ls),$
                                         net_position(indep_ls),$
                                         position_order)
         pfit_slope    = robust_poly_fit(ordernum(indep_ls),$
                                         net_slope(indep_ls),$
                                         slope_order)
      endif else begin
         pfit_fwhm     = poly_fit(ordernum(indep_ls),net_fwhm(indep_ls),$
                                  fwhm_order)
         pfit_position = poly_fit(ordernum(indep_ls),net_position(indep_ls),$
                                  position_order)
         pfit_slope    = poly_fit(ordernum(indep_ls),net_slope(indep_ls),$
                                         slope_order)  
      endelse
      ; Reconstruct total fitted values
      fwhm_fit     = fwhm_hot
      position_fit = position_hot
      slope_fit    = slope_hot
      for k=0,fwhm_order do $
         fwhm_fit = fwhm_fit + pfit_fwhm(k)*(ordernum^k)
      for k=0,position_order do $
         position_fit = position_fit + pfit_position(k)*(ordernum^k)
      for k=0,slope_order do $
         slope_fit = slope_fit + pfit_slope(k)*(ordernum^k)
   endif else begin
      fwhm_fit     = 0.*ordernum
      position_fit = 0.*ordernum
      slope_fit    = 0.*ordernum
   endelse

   ; For the first order, start with no estimates for the object
   ; profile FWHM.  For subsequent orders that are being independently
   ; traced and fit, start with a FWHM estimate from the fiducial
   ; order.  For orders that are not to be independently fit, use
   ; extrapolated values for the object position, object FWHM, and trace
   ; slope from orders for which independent fits have been made.
   if (index eq fid_index) then begin
      if keyword_set(object_xy) then begin
         slitpos = slit(object_xy(0),object_xy(1))
      endif else begin
         slitpos = 0   ; Return fitted slit position
      endelse
      fwhm    = 0   ; FWHM estimate
      slope   = 0
      force   = 0
      independent_fit_made(index) = 1
   endif else begin
      if (do_independent_fit(index) eq 1) then begin
         slitpos = 0
         fwhm    = fwhm_list(fid_index)
         slope   = 0
         force   = 0
         independent_fit_made(index) = 1
      endif else begin
         ; Use fitted values
         slitpos = position_fit(index)
         fwhm    = fwhm_fit(index)
         slope   = slope_fit(index)
         force   = 1
         print,'Using fitted values: Position = '+strtrim(slitpos,2)
         print,'                     FWHM     = '+strtrim(fwhm,2)
         print,'                     Slope    = '+strtrim(slope,2)
         print,' '
      endelse
   endelse
   mean_shift = 0

   ; If using an estimate for the slope of the object trace, adjust
   ; slit grid before passing to longslit_reduce.  If not, trace
   ; the object.
   temp_slit = slit
   if (slope ne 0) then begin
      slit_correction = slope*findgen(ysz)
      for j=0,ysz-1 do temp_slit(*,j) = slit(*,j) - slit_correction(j)
   endif

   ; If using fitted values, make sure tracing is turned off in
   ; longslit_reduce.
   if force then begin
      do_trace = -1
   endif else begin
      ; Default is set by the value of traceobj in the
      ; longslit_reduce.inc file, unless the /notrace keyword is set.
      if keyword_set(notrace) then do_trace = -1 else do_trace = 0
   endelse

   ; Pass header to longslit_reduce
   temp_hdr = hobj

   ; Set gain and read noise
   case ccdname(0) of
      'blue': begin
                 gain    = 0.47  ; e-/ADU
                 rdnoise = 2.00  ; e- 
              end
      'red' : begin
                 gain    = 1.00  ; e-/ADU
                 rdnoise = 3.80  ; 3.50  ; e- (Measured value 10/2007)
              end
   endcase

   ; Set approximate slit positions
   temp_ordercenter = ordercenter(index)
   temp_slitmin     = orderedge1(index)
   temp_slitmax     = orderedge2(index)

   ; Mask a specified region along the slit?
   if apply_mask then $
      objmask = ordercenter(index) + [mask_offset_lo,mask_offset_hi] else $
      objmask = 0
   
   ; Set break point spacing according to approximate dispersion
   this_order   = where(slit ge temp_slitmin and slit le temp_slitmax and $
                        wave gt 100,n_this_order)
   rough_disp   = (max(wave(this_order))-min(wave(this_order))) / $
                  (max(yarr(this_order))-min(yarr(this_order))) 
   bk_space     = 3*rough_disp
   min_bk_space = 0.7*rough_disp
   print,'Setting break point spacing = '+strtrim(bk_space,2)
   print,'    min break point spacing = '+strtrim(min_bk_space,2)

   ; If necessary, override region of the chip in which to locate
   ; object.  This is useful for tracing standards in non-fiducial
   ; orders.
   trace_range = [0,ysz-1-128]

   ; Do extraction in wavelength bins of constant velocity width
   log_lin = 1

   ; Reduce order.  Longslit_reduce will apply heliocentric
   ; velocity correction only to input wavelengths.

   longslit_reduce,obj,wave,temp_slit,temp_nosky,temp_noskyvar,$
      temp_wavefit,temp_skymodel,hdr=temp_hdr,/mike,/nobias,$
      flatfield=flat,flatvar=flat_var,gn=gain,rdnoise=rdnoise,$
      fwhm=fwhm,slitpos=slitpos,force=force,trace_obj=do_trace,$
      meanshift=mean_shift,outfile='temp_mike',$
      tracerange=trace_range,subrange=[temp_slitmin,temp_slitmax],$
      findrange=find_range,objmask=objmask,waveref=wave,$
      bkspace=bk_space,min_bkspace=min_bk_space,/set_bkpts,$
      loglin=log_lin,_extra=ex

   ; Check - Did longslit_reduce run properly?
   if (n_elements(temp_nosky) eq 1 or $
       n_elements(temp_noskyvar) eq 1 or $
       n_elements(temp_wavefit) eq 1 or $
       n_elements(temp_skymodel) eq 1) $
      then longslit_ok = 0 else longslit_ok = 1

   if longslit_ok then begin

      ; Store the section of the 2D reduced frames for this order.
      order_ls = where(slit ge temp_slitmin and slit le temp_slitmax)
      nosky(order_ls)    = temp_nosky(order_ls)
      noskyvar(order_ls) = temp_noskyvar(order_ls)
      wavefit(order_ls)  = temp_wavefit(order_ls)
      skymodel(order_ls) = temp_skymodel(order_ls)
      
      ; Read the new spectrum back in and store in the large array.
      checkexists = findfile('temp_mike_spec.fits',count=n_specfiles)
      checkexists = findfile('temp_mike_err.fits',count=n_errfiles)
      if not(n_specfiles eq 0 or n_errfiles eq 0) then begin
         spec = rflam('temp_mike_spec.fits',hspec,wspec)
         err  = rflam('temp_mike_err.fits',herr,werr)
         n_pix = n_elements(spec)
         big_wave(0:n_pix-1,index) = wspec
         big_spec(0:n_pix-1,index) = spec
         big_err(0:n_pix-1,index)  = err
      endif
      
      ; Record the median wavelength, profile FWHM, slit position,
      ; offset from the standard star, and mean wavelength shift
      ; for this order.
      medwave_list(index)   = median(wspec,/even)
      fwhm_list(index)      = fwhm
      position_list(index)  = slitpos
      meanshift_list(index) = mean_shift

      ; If a trace was made, read in and store trace information
      if do_trace then begin
         trace_templ = {   $
                   version: 1.0, $    
                   datastart: 0L, $   
                   delimiter: ' ', $
                   missingvalue: -99.0, $
                   commentsymbol: '#', $
                   fieldcount: 1L, $
                   fieldtypes: 4L, $
                   fieldnames: 'value', $
                   fieldlocations: 0L, $
                   fieldgroups: [0L] $
                 }
         trace_data = read_ascii('temp_mike_trace.dat',template=trace_templ)
         ; The first numbers are the polynomial fit to the object trace.  The
         ; last number is the object centroid, which should be nearly identical
         ; to the zeroeth term in the trace coefficients.
         all_numbers  = trace_data.value
         n_numbers    = n_elements(all_numbers)
         temp_trace_coeffs = all_numbers(0:n_numbers-2)
         n_coeffs     = n_elements(temp_trace_coeffs)
         trace_coeffs(0:n_coeffs-1,index) = temp_trace_coeffs
      endif else begin
         ; Record fitted values
         trace_coeffs(0:1,index) = [slitpos,slope]
      endelse

   endif

   ; Proceed to next order, or quit.
   if (index_step eq red_direction) then begin
      index = index + index_step
      if (index gt n_orders-1 or index lt 0) then begin
         index_step = blue_direction
         index      = fid_index + index_step
         if (index gt n_orders-1 or index lt 0) then done = 1
      endif
   endif else begin
      index = index + index_step
      if (index gt n_orders-1 or index lt 0) then done = 1
   endelse

endwhile

; Save the 2D reduced arrays.
hout = hobj
get_date,date,/timetag
sxaddhist,'Reduced with MIKE_REDUCE '+date+' (UTC)',hout
sxaddpar,hout,'ORIGEXP',filename(object),'Filename of original exposure.'
sxaddpar,hout,'WAVEGRID',filename(wavefile),'Wavelength-proxy grid'
sxaddpar,hout,'SLITGRID',filename(slitfile),'Slit-position-proxy grid'
sxaddpar,hout,'BZERO',0.0
if keyword_set(arc) then $
   sxaddpar,hout,'ARC',filename(arc),'Comparison lamp'
if keyword_set(flatfield) then $
   sxaddpar,hout,'FLATFLD',filename(flatfield),'Flat field'
if keyword_set(flatvar) then $
   sxaddpar,hout,'FLATVAR',filename(flatvar),'Flat field variance'
writefits,root+'_sub.fits',nosky,hout
writefits,root+'_var.fits',noskyvar,hout
writefits,root+'_sky.fits',skymodel,hout
writefits,root+'_wav.fits',wavefit,hout

; Remove extra zeros and save the big spec and error arrays.
nonzero_ls  = where(total(big_wave,2) gt 0)
max_nonzero = max(nonzero_ls)
big_wave    = big_wave(0:max_nonzero-1,*)
big_spec    = float(big_spec(0:max_nonzero-1,*))
big_err     = float(big_err(0:max_nonzero-1,*))
writefits,root+'_wspec.fits',big_wave,hout
writefits,root+'_spec.fits',big_spec,hout
writefits,root+'_err.fits',big_err,hout

; Output object profile and wavelength calibration information.
openw,profile_lun,root+'_profile.dat',/get_lun
for i=0,n_orders-1 do $
   printf,profile_lun,ordernum(i),medwave_list(i),fwhm_list(i),$
          position_list(i),meanshift_list(i),$
          format='(f8.2,f14.3,3(f12.4))'
free_lun,profile_lun

; Output trace information
openw,trace_lun,root+'_trace.dat',/get_lun
for i=0,n_orders-1 do printf,trace_lun,ordernum(i),trace_coeffs(*,i),$
                             format='(6(f13.6))'
free_lun,trace_lun

; Final fit to object profile and trace parameters (for evaluating
; the fits post-processing)
indep_ls = where(independent_fit_made eq 1 and medwave_list le 8900,n_indep,$
                 complement=fitted_ls,ncomp=n_fitted)
if (n_indep ge 1) then begin
   ; Subtract high-order terms
   net_fwhm     = fwhm_list         - fwhm_hot
   net_position = position_list     - position_hot
   net_slope    = trace_coeffs(1,*) - slope_hot
   ; Fit residuals
   if (n_indep ge n_for_linear) then begin
      pfit_fwhm     = robust_poly_fit(ordernum(indep_ls),$
                                      net_fwhm(indep_ls),$
                                      fwhm_order)
      pfit_position = robust_poly_fit(ordernum(indep_ls),$
                                      net_position(indep_ls),$
                                      position_order)
      pfit_slope    = robust_poly_fit(ordernum(indep_ls),$
                                      net_slope(indep_ls),$
                                      slope_order)
   endif else begin
      pfit_fwhm     = poly_fit(ordernum(indep_ls),net_fwhm(indep_ls),$
                               fwhm_order)
      pfit_position = poly_fit(ordernum(indep_ls),net_position(indep_ls),$
                               position_order)
      pfit_slope    = poly_fit(ordernum(indep_ls),net_slope(indep_ls),$
                                      slope_order)  
   endelse
   ; Reconstruct total fitted values
   fwhm_fit     = fwhm_hot
   position_fit = position_hot
   slope_fit    = slope_hot
   for k=0,fwhm_order do $
      fwhm_fit = fwhm_fit + pfit_fwhm(k)*(ordernum^k)
   for k=0,position_order do $
      position_fit = position_fit + pfit_position(k)*(ordernum^k)
   for k=0,slope_order do $
      slope_fit = slope_fit + pfit_slope(k)*(ordernum^k)
endif else begin
   fwhm_fit     = 0.*ordernum
   position_fit = 0.*ordernum
   slope_fit    = 0.*ordernum
endelse

; Plot profile and trace information
fwhm_plotmin     = min(fwhm_list) < min(fwhm_fit)
fwhm_plotmax     = max(fwhm_list) > max(fwhm_fit)
position_plotmin = min(position_list) < min(position_fit)
position_plotmax = max(position_list) > max(position_fit)
slope_plotmin    = min(trace_coeffs(1,*)) < min(slope_fit)
slope_plotmax    = max(trace_coeffs(1,*)) > max(slope_fit)
ps_openfile,root+'_trace',/portrait
   !p.multi    = [0,2,3]
   !p.charsize = 1.2
   !x.charsize = 1.2
   !y.charsize = 1.2
   !x.range    = [min(ordernum)-0.5,max(ordernum)+0.5]
   !x.style    = 1
   ; FWHM
   plot,ordernum,fwhm_list,psym=1,title='FWHM',$
        yr=[fwhm_plotmin,fwhm_plotmax]
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),fwhm_list(fitted_ls),psym=6
   oplot,ordernum,fwhm_fit
   plot,ordernum,fwhm_list-fwhm_fit,psym=1,title='FWHM Residuals'
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),(fwhm_list-fwhm_fit)(fitted_ls),psym=6
   ; Position
   plot,ordernum,position_list,psym=1,title='Position',$
        yr=[position_plotmin,position_plotmax]
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),position_list(fitted_ls),psym=6
   oplot,ordernum,position_fit
   plot,ordernum,position_list-position_fit,psym=1,title='Position Residuals'
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),(position_list-position_fit)(fitted_ls),psym=6
   ; Slope
   plot,ordernum,trace_coeffs(1,*),psym=1,title='Slope',$
        yr=[slope_plotmin,slope_plotmax]
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),trace_coeffs(1,fitted_ls),psym=6
   oplot,ordernum,slope_fit
   plot,ordernum,trace_coeffs(1,*)-slope_fit,psym=1,title='Slope Residuals'
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),(trace_coeffs(1,*)-slope_fit)(fitted_ls),psym=6
ps_closefile

;; Remove temporary files.
print,'Removing temporary files...'
spawn,'"rm" temp_mike_err.fits'
spawn,'"rm" temp_mike_profile.ps'
spawn,'"rm" temp_mike_sky.fits'
spawn,'"rm" temp_mike_spec.fits'
spawn,'"rm" temp_mike_spec.ps'
spawn,'"rm" temp_mike_sub.fits'
spawn,'"rm" temp_mike_trace.ps'
spawn,'"rm" temp_mike_trace.dat'
spawn,'"rm" temp_mike_var.fits'
spawn,'"rm" temp_mike_wav.fits'

print,'Done.'

; Close log
journal

; Restore plot settings
cleanplot,/silent

return
end
