Pro mike_slitgrid,stdfile,slitgrid,orderlist,flatfile=flatfile,$
                  oldgrid=oldgrid,blue=blue,red=red

;+--------------------------------------------------------------------------
;
; MIKE_SLITGRID      2/2007
;
; Trace orders in a bright standard star exposure to create a map of slit
; positions.  The slit position at a pixel is the x-intercept of the
; curve of constant position along the slit that would pass through
; the center of that pixel.
;
; Standard star should be at low airmass (less than 1.1)
;
; Must set either the /blue or /red keyword.
;
; Assumes illuminated region is 1024x2048 (2x2 binning)
;
; NOTE: Rather than identyfying the order edges by hand, it may be
;       easier to determine the order positions using the known size
;       of the slit (assuming the projected size doesn't change much
;       across the array).
;
; INPUTS
;         stdfile   - String, FITS filename of standard star exposure
;
; KEYWORDS
;         flatfile  - String, FITS filename of flatfield exposure
;                     (non-diffused).  Used to identify the edges
;                     of each order.
;         oldgrid   - String, FITS filename of an already made
;                     slitgrid.  If set, program will skip creating a
;                     new grid and use this slitgrid to find the order
;                     edges (if the flatfile keyword is set).
;         blue      - Blue CCD
;         red       - Red CCD
;
; OUTPUTS
;         slitgrid  - Name of FITS file to which to write slit grid
;         orderlist - Name of ASCII file to which to write list of
;                     order centers and edges.  Output only if the
;                     flatfile keyword is set.
;
; HISTORY
;         Written 2/14/2007 GDB
;         Modified to trace standard star 5/10/2007
;         Modified way in which order centers are calculated 9/16/2011
;---------------------------------------------------------------------------

if (n_params() lt 2) then begin
 print,'CALLING SEQUENCE: '
 print,'   mike_slitgrid,stdfile,slitgrid,orderlist,flatfile=flatfile,'
 print,'                 oldgrid=oldgrid,/blue,/red'
 return
endif

device,decomp=0
loadct,39

;;; Is this for the blue side or the red side?
if keyword_set(blue) then begin
   cut_ymin     = 600
   cut_xmin     = 10
   cut_xmax     = 1000
   npoly_coeffs = 3
   ccd_name     = 'blue'
endif else begin
   if keyword_set(red) then begin
      cut_ymin     = 1100
      cut_xmin     = 10
      cut_xmax     = 1000
      npoly_coeffs = 3
      ccd_name     = 'red'
   endif else begin
      print,'Must set either the /blue or /red keyword.'
      return
   endelse
endelse

;;; Assume 2x2 binning
illum_xsz = 1024
illum_ysz = 2048

if not(keyword_set(oldgrid)) then begin

   ;;; Create a new slit grid from a standard star exposure
   
   ;;; Read in standard star exposure
   std  = mike_subbias(readfits(stdfile,hstd))
   sz   = size(std)
   xsz  = sz(1)
   ysz  = sz(2)
   
   ;;; Median cut through the array
   cut = median(std(*,cut_ymin:cut_ymin+100),dim=2)
   
   ;;; Locate peaks.  Avoid edges of illuminated region.
   first_peak = 1
   med_cut    = median(cut)
   for i=cut_xmin,cut_xmax do begin
      if (cut(i) eq max(cut(i-10:i+10)) and $
          cut(i) gt med_cut) then begin
         if (first_peak) then begin
            peak_list  = i
            first_peak = 0
         endif else begin
            peak_list  = [peak_list,i]
         endelse
      endif
   endfor
   n_peak = n_elements(peak_list)
   
   ; Plot peak locations
   window,9,xs=1000,ys=500,title='Peak Locations'
   plot,cut,psym=10
   for i=0,n_peak-1 do oplot,replicate(peak_list(i),2),[-1e6,1e6],color=250
   
   ; Determine the x and y ranges over which to trace flats.
   xrange = [1,illum_xsz-2]
   yrange = [0,illum_ysz-1]
   
   ; Set trace fit polynomial
   npoly_trace = 5  ; A bright standard is helpful!
   
   ; Big arrays to hold fit coefficients and uncertainties 
   ; for all traces.
   trcoeffs     = fltarr(npoly_trace+1,n_peak)
   sig_trcoeffs = fltarr(npoly_trace+1,n_peak)
   
   ; List of std deviations from trace fits.  Use to flag bad traces.
   xerr_list = fltarr(n_peak)
   
   ; Trace and fit each peak
   print,'Tracing...'
   window,10,xs=1000,ys=600,title='Traces'
   !p.multi = [0,ceil(sqrt(n_peak)),ceil(sqrt(n_peak))]
   for j=0,n_peak-1 do begin
      traceline,std,peak_list(j),cut_ymin,xtr,ytr,/ydir,merit=merit,$
                median=11,range=yrange,bound=xrange,nside=10,/noplot,/bygauss
      temp_trcoeffs = robust_poly_fit(ytr,xtr,npoly_trace,xfit,xerr)
      ; Reject outliers and refit to get uncertainties in
      ; coefficients.
      ls = where(abs(xtr-xfit) le 3*xerr)
      trcoeffs(*,j) = poly_fit(ytr(ls),xtr(ls),npoly_trace,$
                               sigma=sig,yerror=xerr,yfit=xfit)
      sig_trcoeffs(*,j) = sig
      ; Plot residuals
      plot,xtr(ls)-xfit,ytr(ls),psym=3,/xstyle,/ystyle
      ; Record std dev from fit
      xerr_list(j) = xerr
   endfor
   
   ; Array to hold fit coefficients for fits to the trace coefficients.
   ; Must have second dimension = 4 to work with xy2a.pro or xy2b.pro.
   fitcoeffs = fltarr(npoly_trace,4)
   ; Fit coefficients
   chan,11
   !p.multi = [0,1,1]
   ; Default is to include all traces where the standard deviation from 
   ; the fit is not spuriously small
   include = 1 + fltarr(n_peak)
   bad_ls  = where(xerr_list lt 1e-4,n_bad)
   if (n_bad ne 0) then include(bad_ls) = 0
   for j=1,npoly_trace do begin
      print,'Fitting trace coefficients...'
      chan,11
      c0 = trcoeffs(0,*)
      cj = trcoeffs(j,*)
      ls = sort(c0)
      temp_fitcoeffs = inter_poly_fit(c0(ls),cj(ls),npoly_coeffs,$
                                      include=include,$
                                      measure_errors=sig_trcoeffs(j,*))
      fitcoeffs(j-1,0:npoly_coeffs) = temp_fitcoeffs
   endfor
   ; Compute slit grid for the array
   print,'Computing slit grid...'
   slit = xy2a(findgen(xsz),findgen(ysz),fitcoeffs)
   slit = float(slit)
   
   ;;; Write slit grid to FITS file
   sxaddpar,hstd,'BZERO',0.0
   writefits,slitgrid,slit,hstd

endif else begin

   ;;; Use an already made slit grid
   slit = readfits(oldgrid,hslit)
   sz   = size(slit)
   xsz  = sz(1)
   ysz  = sz(2)

endelse


;;; Identify order edges (optional)
if keyword_set(flatfile) then begin
   flat = mike_subbias(readfits(flatfile,hflat))
   slit_displaymin = min(slit(0:illum_xsz-1,cut_ymin))
   slit_displaymax = max(slit(0:illum_xsz-1,cut_ymin))
   display_step    = 100.
   n_display       = ceil((slit_displaymax-slit_displaymin) / $
                          float(display_step))
   first_orderedge = 1
   done            = 0
   window,11,xs=1000,ys=500
   print,'Left-click once on each edge of each order.  Ignore partial orders.'
   print,'Right-click when done with each set.'
   slit_slice = slit(0:illum_xsz-1,300:1700)
   flat_slice = flat(0:illum_xsz-1,300:1700)
   for i=0,n_display-1 do begin
      x_min = slit_displaymin + i*display_step - 10
      x_max = slit_displaymin + (i+1)*display_step + 10
      y_min = 0
      plot_ls = where(slit_slice ge x_min and slit_slice le x_max)
      y_max = cutvalue(flat_slice(plot_ls),99.9)
      plot,slit_slice(plot_ls),flat_slice(plot_ls),psym=3,xr=[x_min,x_max],/xs,$
           yr=[y_min,y_max]
      if not(first_orderedge) then begin
         for j=0,n_elements(orderedge_list)-1 do $
            oplot,[orderedge_list(j),orderedge_list(j)],[-1e6,1e6],$
                  linestyle=1,color=150
      endif   
      getpoints,temp_orderedge,/plotx,color=150
      if (total(temp_orderedge) ne -1) then begin
         if first_orderedge then begin
            orderedge_list  = temp_orderedge
            first_orderedge = 0
         endif else begin
            orderedge_list  = [orderedge_list,temp_orderedge]
         endelse
      endif
   endfor
   orderedge_list = orderedge_list(sort(orderedge_list))
   
   ;;; Compute center of each order
   n_orderedge = n_elements(orderedge_list)
   if (n_orderedge mod 2 ne 0) then begin
      print,'MIKE_SLITGRID: Error - There should be an even number of order' 
      print,'               edges marked (two for each order).  Exiting without'
      print,'               writing order list.'
      return
   endif
   n_order        = n_orderedge / 2
   leftedge_list  = orderedge_list(2*lindgen(n_order))
   rightedge_list = orderedge_list(1+2*lindgen(n_order))
   ;center_list    = (leftedge_list + rightedge_list) / 2.
   ; Fit a smooth function to the order midpoints
   midpoint_list  = (leftedge_list + rightedge_list) / 2.
   center_polyorder = floor(n_order/2.) < 5
   ordernum         = findgen(n_order)
   center_pfit      = robust_poly_fit(ordernum,midpoint_list,center_polyorder)
   center_list      = 0.*midpoint_list
   for i=0,center_polyorder do center_list = center_list + $
                                             center_pfit(i)*(ordernum^i)
   
   ;;; Write order info to file
   print,'Writing order info to file...'
   openw,unit,orderlist,/get_lun
   printf,unit,'#  CCD   Order    Center   Left edge  Right edge'
   for i=0,n_order-1 do $
      printf,unit,ccd_name,i+1,center_list(i),leftedge_list(i),$
                  rightedge_list(i),format='(a6,i6,3(f12.3))'
   free_lun,unit
endif

return
end
