Function mike_subbias,raw

;+---------------------------------------------------------------------
;
; MIKE_SUBBIAS     2/2007 GDB
; 
; Subtract the bias from a raw MIKE frame by computing the mean bias
; level in the overscan region (both in x and y).  Errors are probably
; < 1 count for each row or column, and no significant noise is added
; to the array.
;
; Assumes the last 128 rows and 128 columns are overscan (should be OK
; for any binning).
;
; INPUTS
;          raw      - Raw 2D array, can be a named variable or the name
;                     of a FITS file.
;
; OUTPUTS
;          <result> - 2D bias-subtracted array.
;
; HISTORY
;          Adapted from LDSS3_SUBBIAS 2/14/2007 GDB
;---------------------------------------------------------------------

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE: result = mike_subbias(raw)'
   return,-1
endif

; Read in FITS file, if necessary
if (size(raw,/tname) eq 'STRING') then begin
   checkexists = findfile(raw,count=nfiles)
   if (nfiles eq 0) then begin
      print,'MIKE_SUBBIAS: Raw file not found.'
      return,-1
   endif else begin
       rawframe = readfits(raw,hrawframe)
       fromfitsfile = 1 
   endelse
endif else rawframe = raw

; Size of illuminated region + overscan
sz  = size(rawframe)
xsz = sz(1)
ysz = sz(2)

; Overscan size
oscan_xsz = 128
oscan_ysz = 128

; Bias-subtracted frame
subframe = 1.*rawframe  ; Promote to floating point

; Measure the mean bias level across the overscan region in x and
; subtract from each row.  Must do bad pixel rejection (at least for
; the y-overscan).
for i=0,ysz-1 do begin
   tmp_oscan     = subframe(xsz-oscan_xsz:xsz-1,i)
   med_level     = median(tmp_oscan)
   sig_level     = robust_sigma(tmp_oscan)
   good_ls       = where(abs(tmp_oscan-med_level) le 4*sig_level)
   mean_level    = mean(tmp_oscan(good_ls))
   subframe(*,i) = subframe(*,i) - mean_level
endfor

; Measure the mean bias level across the overscan region in y and
; subtract from each column.  Must do bad pixel rejection (at least for
; the y-overscan).
for i=0,xsz-1 do begin
   tmp_oscan     = subframe(i,ysz-oscan_ysz:ysz-1)
   med_level     = median(tmp_oscan)
   sig_level     = robust_sigma(tmp_oscan)
   good_ls       = where(abs(tmp_oscan-med_level) lt 4*sig_level)
   mean_level    = mean(tmp_oscan(good_ls))
   subframe(i,*) = subframe(i,*) - mean_level
endfor


return,subframe
end
