Pro mike_wavegrid,arcfile,orderlist,slitgrid,wavegrid

;+----------------------------------------------------------------------------
;
; MIKE_WAVEGRID
;
; Trace MIKE arc lines to create a map of wavelength-proxies for each
; order.
;
; Assumes illuminated region is 1024x2048 (2x2 binning).
;
; INPUTS
;         arcfile      - FITS filename of arc lamp exposure
;         orderlist    - Name of ASCII file containing the order
;                        information [CCD, order #, center, left edge,
;                        right edge], as output by mike_slitgrid.pro.
;         slitgrid     - FITS filename of slit position map created by
;                        mike_slitgrid.pro.
;
; KEYWORDS
;
; OUTPUTS
;         wavegrid     - FITS filename of uncalibrated wavelength grid.
;
; HISTORY
;         Written 2/17/2007 GDB
;-----------------------------------------------------------------------------

if (n_params() lt 4) then begin
   print,'CALLING SEQUENCE: '
   print,'       mike_wavegrid,arcfile,orderlist,slitgrid,wavegrid'
   return
endif

; Plot in color
device,decomp=0
loadct,39

; Read in center slit positions and CCD names
order_templ =  {   $
          version: 1.0, $
          datastart: 0, $
          delimiter: ' ', $        ; Fields separated by white space
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 5, $
          fieldtypes: [7,4,4,4,4], $
          fieldnames: ['ccd','ordernum','center','left','right'], $
          fieldlocations: [2,11,18,30,42], $
          fieldgroups: [0,1,2,3,4] $
        }
orderinfo    = read_ascii(orderlist,template=order_templ)
tmp_ordernum    = orderinfo.ordernum
tmp_ordercenter = orderinfo.center
tmp_orderedge1  = orderinfo.left
tmp_orderedge2  = orderinfo.right
; Sort orders in order of increasing slit position
order_order = sort(tmp_ordercenter)
ordernum    = tmp_ordernum(order_order)
ordercenter = tmp_ordercenter(order_order)
orderedge1  = tmp_orderedge1(order_order)
orderedge2  = tmp_orderedge2(order_order)
n_orders    = n_elements(ordercenter)

; Read in arc and slit grid
arc  = mike_subbias(readfits(arcfile,harc))
slit = readfits(slitgrid,hslit)

; Set order for trace fit polynomial
npoly_trace = 1

; Set order for coefficient fit polynomial
npoly_coeffs = 1

; Big array to hold fit coefficients for the trace fit coefficients,
; has to have 2nd dimension = 4 to work with xy2b.pro.
fitcoeffs     = fltarr(npoly_trace,4,n_orders)
sig_fitcoeffs = fltarr(npoly_trace,4,n_orders)

; Illuminated region (currently set for 2x2 binning).
illum_xsz = 1024.
illum_ysz = 2048.

; Arrays of x and y-values
sz    = size(arc)
xsz   = sz(1)
ysz   = sz(2)
yarr  = lonarr(xsz,ysz)
for i=0,ysz-1 do yarr(*,i) = i
xarr  = lonarr(xsz,ysz)
for i=0,xsz-1 do xarr(i,*) = i
xlist = dindgen(xsz)
ylist = dindgen(ysz)

; Loop through orders.  Select and trace arc lines.

for i=0,n_orders-1 do begin

   ; Select pixels within the current order
   tmp_ls = where(slit ge orderedge1(i) and slit le orderedge2(i) and $
                  yarr le illum_ysz-1,n_tmp)
   if (n_tmp eq 0) then begin
      print,'Error - Order '+strtrim(ordernum(i),2)+$
            ' does not contain any pixels.'
stop
      return
   endif

   ; Interpolate counts near the center of the slit onto a regular grid in y
   center_ls = where(slit ge ordercenter(i)-1 and slit le ordercenter(i)+1 and $
                     yarr le illum_ysz-1)
   yvals  = yarr(center_ls)
   xvals  = xarr(center_ls)
   avals  = arc(center_ls)
   y_ord  = sort(yvals)
   counts = interpol(avals(y_ord),yvals(y_ord),ylist)

   ; Choose arc lines.  Avoid lines near the edge of the array
   ypeak    = -1
   first_peak = 1
   for j=10,illum_ysz-10 do begin
      med_ymin = (0 > (ylist(j) - 20)) < (max(yvals) - 41)
      med_ymax = med_ymin + 40
      if (counts(j) eq max(counts(j-5:j+5)) and $
          counts(j) gt median(counts(med_ymin:med_ymax)) + $
                       7*robust_sigma(counts(med_ymin:med_ymax)) and $
          counts(j) lt 40000. and $             ; Avoid saturated lines
          counts(j) ne counts(j-1)) then begin  ; Avoid double-counting
         if first_peak then begin
            ypeak      = [ylist(j)]
            first_peak = 0
         endif else begin
            ypeak      = [ypeak,ylist(j)]
         endelse
      endif
   endfor
   n_peak = n_elements(ypeak)

   ; Show located lines
   if (i eq 0) then window,9,xsize=1000,ysize=500 else chan,9
   !p.multi = [0,1,1]
   plot,yvals,arc(center_ls),psym=3,/xs,title='Order '+strtrim(ordernum(i),2)
   oplot,ylist,counts,psym=10,color=250
   for j=0,n_peak-1 do oplot,[ypeak(j),ypeak(j)],[-1e6,1e6],color=150,$
                             linestyle=1

   ; Get x-positions at the peak points
   xpeak = fltarr(n_peak)
   for j=0,n_peak-1 do xpeak(j) = mean(xvals(where(yvals ge ypeak(j)-1 and $
                                                   yvals le ypeak(j)+1)))

   ; Trace arc and fit arc lines
   print,'Tracing lines...'
   if (i eq 0) then window,10,xs=1000,ys=800 else chan,10
   ntiles2 = floor(sqrt(n_peak))
   ntiles1 = ceil((1.*n_peak)/ntiles2)
   !p.multi=[0,ntiles1,ntiles2]
   trcoeffs     = fltarr(npoly_trace+1,n_peak)
   sig_trcoeffs = fltarr(npoly_trace+1,n_peak)
   for j=0,n_peak-1 do begin
      traceline,arc,xpeak(j),ypeak(j),xtr,ytr,slittr,/xdir,/bygauss,$
                merit=merit,median=1,nside=5,coord=slit,cbound=[orderedge1(i),$
                orderedge2(i)],/noplot
      plot,xtr,ytr,psym=3,/xs,/ys,title='y = '+strtrim(ypeak(j),2)
      ; Reject bad trace points.  Also attempt to avoid reflected 
      ; lines, which would take up only part of the slit.
      goodls = where(merit ne 0,n_good)
      if (n_good gt 0) then begin
         if (n_good gt npoly_trace+1 and $
             min(slittr(goodls)) le orderedge1(i)+2 and $
             max(slittr(goodls)) ge orderedge2(i)-2 and $
             n_good gt 0.8*n_elements(xtr)) then begin
            good_xtr = xtr(goodls)
            good_ytr = ytr(goodls)
            temp_trcoeffs = robust_poly_fit(good_xtr,good_ytr,npoly_trace,$
                                            yfit,yerr)
            ; Reject outliers and refit to get uncertainties in coefficients.
            ls = where(abs(good_ytr-yfit) le 3*yerr)
            trcoeffs(*,j) = poly_fit(good_xtr(ls),good_ytr(ls),npoly_trace,$
                                     sigma=sig,yerror=yerr,yfit=yfit)
            if (yerr lt 0.5) then begin
               sig_trcoeffs(*,j) = sig
               oplot,good_xtr(ls),yfit,color=250
            endif else begin
               trcoeffs(*,j) = -99
               oplot,good_xtr(ls),yfit,color=50
            endelse
         endif else begin
            trcoeffs(*,j) = -99
         endelse
      endif else begin
         trcoeffs(*,j) = -99
      endelse
   endfor

   ; Fit coefficients
   trcoeff0 = trcoeffs(0,*)
   trcoeff1 = trcoeffs(1,*)

   ; Reject any lines where the intercept seems unreasonable or where
   ; the linear slope is > 1.5 or < -1.5.
   min_trcoeff0 = -500
   max_trcoeff0 = 2550
   goodls = where(trcoeff0 ge min_trcoeff0 and trcoeff0 le max_trcoeff0 and $
                  abs(trcoeff1) lt 1.5 and trcoeff0 ne -99,n_good)
   trcoeff0 = trcoeff0(goodls)
   ; If there is more than one good line, then fit trace coefficients
   if (n_good gt 1) then begin
      for j=1,npoly_trace do begin
         trcoeffj         = trcoeffs(j,goodls)
         sig_trcoeffj     = sig_trcoeffs(j,goodls)
         converged        = 0
         old_trcoeffj_fit = 0
         include_ls       = lindgen(n_good)
         n_include        = n_good
         iter             = 0
         max_iter         = 10
         ; Iterate fit
         while not(converged) do begin
            iter = iter + 1
            temp_fitcoeffs = poly_fit(trcoeff0(include_ls),$
                                      trcoeffj(include_ls),$
                                      npoly_coeffs,$
                                      measure_errors=sig_trcoeffj(include_ls),$
                                      sig=temp_sig_fitcoeffs,$
                                      yfit=trcoeffj_fit_include)
            stddev_trcoeffj_fit = $
                   robust_sigma(trcoeffj(include_ls)-trcoeffj_fit_include)
            ; Check that robust_sigma has returned a sensible result
            if (stddev_trcoeffj_fit eq -1 or $
                stddev_trcoeffj_fit lt 0.5*median(sig_trcoeffj(include_ls))) $
               then stddev_trcoeffj_fit = $
                            stddev(trcoeffj(include_ls)-trcoeffj_fit_include)
            trcoeffj_fit   = 0.*trcoeff0
            for k=0,npoly_coeffs do $
               trcoeffj_fit = trcoeffj_fit + (trcoeff0^k)*temp_fitcoeffs(k)
            if (max(abs(old_trcoeffj_fit - trcoeffj_fit)) lt 1e-9 or $
                iter ge max_iter or $
                n_elements(trcoeff0) eq 2 or $
                stddev_trcoeffj_fit eq 0) then begin
               converged = 1
               !p.multi = [0,1,1]
               chan,13
               ploterror,trcoeff0,trcoeffj,sig_trcoeffj,psym=4,$
                         title='Order '+strtrim(ordernum(i),2)
               oplot,trcoeff0(include_ls),trcoeffj(include_ls),psym=4,color=250
               oplot,trcoeff0,trcoeffj_fit,color=200
            endif else begin
               trcoeffj_resid = trcoeffj-trcoeffj_fit
               include_ls = $
                   where(abs(trcoeffj_resid) lt 4.*stddev_trcoeffj_fit and $
                         abs(trcoeffj_resid) le 0.04,n_include)
               old_trcoeffj_fit = trcoeffj_fit
            endelse
         endwhile
         fitcoeffs(j-1,0:npoly_coeffs,i) = temp_fitcoeffs
         sig_fitcoeffs(j-1,0:npoly_coeffs,i) = temp_sig_fitcoeffs
      endfor
   endif else begin
      ; Dummy fit coefficents.  Boost errors so that these
      ; points are ignored.
      fitcoeffs(*,*,i)     = 0
      sig_fitcoeffs(*,*,i) = 1e29
   endelse
endfor

; Replace outliers with coefficients modeled across the entire array
npoly_model     = 2
model_fitcoeffs = 0*fitcoeffs
outlier_ls      = lonarr(n_orders)
for i=0,npoly_trace-1 do begin
   for j=0,npoly_coeffs do begin
      pfit = poly_fit(ordernum,fitcoeffs(i,j,*),npoly_model,$
                      measure_errors=sig_fitcoeffs(i,j,*),yfit=yfit)
      yerr = robust_sigma(fitcoeffs(i,j,*)-yfit)
      model_fitcoeffs(i,j,*) = yfit
      tmp_outlier_ls = where(abs(fitcoeffs(i,j,*)-yfit) ge 7*yerr,tmp_n_out)
      if (tmp_n_out ne 0) then outlier_ls(tmp_outlier_ls) = 1         
   endfor
endfor
if (total(outlier_ls) ne 0) then begin
   replace_ls = where(outlier_ls eq 1)
   print,'Replacing coeffients for orders ',$
         strtrim(ordernum(replace_ls),2),' with model coefficients.'
   fitcoeffs(*,*,replace_ls) = model_fitcoeffs(*,*,replace_ls)
endif

; Write fit coefficients to file
openw,unit,'temp_fitcoeffs.dat',/get_lun
for i=0,n_orders-1 do printf,unit,i,fitcoeffs(*,*,i),format='(i6,4(f16.10))'
free_lun,unit

; Create wavelength arrays with the values for each order.

wavearr = dblarr(xsz,ysz)
for i=0,n_orders-1 do begin
   ; Compute slit positions for this order
   case i of
      0:          order_area = $
                         where(slit lt 0.5*(orderedge2(i)+orderedge1(i+1)))
      n_orders-1: order_area = $
                         where(slit ge 0.5*(orderedge2(i-1)+orderedge1(i)))
      else:       order_area = $
                         where(slit ge 0.5*(orderedge2(i-1)+orderedge1(i)) and $
                               slit lt 0.5*(orderedge2(i)+orderedge1(i+1)))
   endcase
   ; Speed up computation by only computing wavelengths over the 
   ; x-range covered by this order.
   order_xmin  = min(xarr(order_area))
   order_xmax  = max(xarr(order_area))
   tmp_xlist   = order_xmin + dindgen(order_xmax-order_xmin+1)
   tmp_wavearr = xy2b(tmp_xlist,ylist,fitcoeffs(*,*,i))
   big_tmp_wavearr = dblarr(xsz,ysz)
   big_tmp_wavearr(order_xmin:order_xmax,*) = tmp_wavearr
   ; Store wavelengths for this order
   wavearr(order_area) = big_tmp_wavearr(order_area)
endfor

; Write wavelength array to FITS file
sxaddpar,harc,'BZERO',0.0
writefits,wavegrid,wavearr,harc

; Restore plot settings
cleanplot,/silent

return
end
