Function cubic_roots,coeffs

;+--------------------------------------------------------------------
;
; CUBIC_ROOTS       
;
; Returns the real roots of a cubic equation of the form
;
;                  0 = c0 + c1*x + c2*x^2 + c3*x^3 .
;
; Follows standard proceedure documented in many places.  Computation
; and results are in double floating point precision.
;
; INPUTS
;         coeffs   - 4-element array of real coefficients
;                    [c0,c1,c2,c3].  c3 should not be 0.
;
; OUTPUTS
;         <result> - Three-element array, if three real roots
;                    exist, or a one-element array, if only one real
;                    root exists.
;
; HISTORY
;         Written 8/4/2004 GDB
;---------------------------------------------------------------------

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE: result = cubic_roots(coeffs,/double)'
   return,-1
endif

; Convert coefficients to double floating point

c0 = 1d0 * coeffs(0)
c1 = 1d0 * coeffs(1)
c2 = 1d0 * coeffs(2)
c3 = 1d0 * coeffs(3)

; Divide out the x^3 coefficient to get equation in the form 
; x^3 + ax^2 + bx + c = 0

a = c2 / c3
b = c1 / c3
c = c0 / c3

; Check if three real roots exist

q = (a^2 - 3*b)/9
r = (2*a^3 - 9*a*b + 27*c)/54

if (r^2 le q^3) then threeroots = 1 else threeroots = 0

; Compute root(s)

if threeroots then begin ; Compute three real roots
   theta = acos(r/sqrt(q^3))
   root1 = -2*sqrt(q)*cos(theta/3) - a/3
   root2 = -2*sqrt(q)*cos((theta + 2*!dpi)/3) - a/3
   root3 = -2*sqrt(q)*cos((theta - 2*!dpi)/3) - a/3
   roots = [root1,root2,root3]
endif else begin         ; Compute sole real root
   signr = r/abs(r)
   g = -1*signr*((abs(r) + sqrt(r^2 - q^3))^(1d0/3))
   if (a eq 0) then h = 0 else h = q/g
   root1 = (g+h) - a/3
   roots = [root1]
endelse

return,roots

end
