Function gauss_interp,x,y,yerr,n,sig,chisq=chisq,wrap=wrap,x0=x0,$
                      interval=interval

;+-----------------------------------------------------------------------
;
; GAUSS_INTERP       8/2007 
;
; Perform a local interpolation at each pixel along a 1D array
; using polynomial fits within a Gaussian weighting window.
;
; INPUTS
;       x       - 1D vector, indeendent variable.  Should be 
;                 evenly spaced.
;       y       - 1D vector, data to be interpolated
;       yerr    - 1D vector of 1-sigma uncertainties in y
;       n       - Order of the polynomial to be used for performing
;                 local fits around each pixel (if unsure, try 1 or
;                 2).
;       sig     - Width of the Gaussian weighting function
;                 to be used, in units of x.  May either be a scalar
;                 of a vector.  If a vector, then the largest width
;                 is used first, and smaller widths are used if the
;                 chi-squared is poor.
;
; KEYWORDS
;       chisq   - Returns the reduced chi-squared values at
;                 each pixel, where the errors have been weighted
;                 by the Gaussian window function.
;       wrap    - Treat x and y as periodic, and perform convolution
;                 at the edged by "wrapping" arrays.  Default is
;                 to use fewer pixels at the edges when computing 
;                 local polynomial fits.
;       x0      - Scalar.  If set, compute the interpolation only at
;                 the pixel nearest to this point.  Program will
;                 return a scalar value.
;       interval - Scalar, interval at which to compute local gaussian
;                  interpolations, in units of x.  Default is to
;                  compute gaussian interpolaton at every pixel.  If
;                  an interval larger than the pixel size is used, the
;                  remainder of the values will be interpolated using
;                  a spline function from those pixels where the gaussian
;                  interpolation has been computed.  For the Lya forest,
;                  intervals up to ~5 km/s give acceptable results.
;
; OUTPUTS
;       <result> - Interpolated version of y, where the value
;                  at each pixel is the value of the local polynomial
;                  fit.
;
; HISTORY
;       Written 7/24/2007 GDB
;       WRAP keyword added 2/26/2009 GDB
;       X0 keyword added 4/29/2009 GDB
;------------------------------------------------------------------------

if (n_params() lt 5) then begin
   print,'CALLING SEQUENCE: '
   print,'      result = gauss_interp(x,y,yerr,n,sig,chisq=chisq,/wrap,'
   print,'                            x0=x0,interval=interval)'
   return,-1
endif

; Check that the error is everywhere non-negative
zero_ls = where(yerr le 0,n_zero)
if (n_zero ne 0) then begin
   print,'Error arrays must be > 0 everywhere.'
   return,-1
endif

; Determine the maximum distance in x to which pixels
; will be included in the local polynomial fit.  Include
; pixels where the Gaussian is > 1% of its max value.
gmin   = 0.01
dx_max = max(sig)*sqrt(2*alog(1./gmin))

; Assume pixels are regularly spaced in x
dx_perpix = x(1) - x(0)
nside     = ceil(dx_max/dx_perpix)
if (nside lt 1) then begin
   print,'GAUSS_INTERP: nside should be >= 1'
   return,-1
endif

; Optional - Wrap edges of the arrays
if keyword_set(wrap) then begin
   x_orig    = x
   y_orig    = y
   yerr_orig = yerr
   npix   = n_elements(x)
   x    = [x(0)-dx_perpix*(1+reverse(findgen(nside))),x,$
           x(npix-1)+dx_perpix*(1+findgen(nside))]
   y    = [y(npix-1-(nside-1):npix-1),y,y(0:nside-1)]
   yerr = [yerr(npix-1-(nside-1):npix-1),yerr,yerr(0:nside-1)]
   ; Number of pixels to omit at the end
   n_edge = nside
endif else begin
   n_edge = 0
endelse

; Arrays to store interpolated values
yinterp = 0.*y
chisq   = 0.*y
npix    = n_elements(x)
sigused = fltarr(npix)  ; Testing

; Gaussian kernel(s).  Largest width first.
sigorder = reverse(sort(sig))
siglist  = sig(sigorder)
nsig    = n_elements(siglist)
kernel  = fltarr(2*nside+1,nsig)
for i=0,nsig-1 do begin
   temp_kernel = gaussker(siglist(i)/dx_perpix,size=2*nside+1)
   kernel(*,i) = temp_kernel / max(temp_kernel)
endfor

; Optional - Compute interpolation only at a single point.  Assume
; points are ordered in x.
if keyword_set(x0) then begin
   i0     = round(interpol(findgen(npix),x,x0))
   istart = i0
   iend   = i0
endif else begin
   istart = 0L
   iend   = long(npix)-2
endelse

; Optional - Compute Gaussian interpolations at regular intervals 
; rather than at every pixel.
if keyword_set(interval) then begin
   istep = interval/dx_perpix
endif else begin
   istep = 1
endelse
nstep = floor((iend-istart)/float(istep) + 1)

; At each pixel, compute an interpolated value based on a local
; polynomial fit, weighting neighboring pixels with a Gaussian.
max_chisq = 3.0
i = istart
while (i le iend) do begin
   ; Select pixels within dx_max of this position
   nleft      = nside < i
   nright     = nside < (npix-1-i)
   ilo        = i-nleft
   ihi        = i+nright
   local_x    = x(ilo:ihi) - mean(x(ilo:ihi))
   local_y    = y(ilo:ihi)
   local_yerr = yerr(ilo:ihi)
   fit_done   = 0
   ; Start with widest Gaussian
   j          = 0
   while not(fit_done) do begin
      local_ker   = kernel(nside-nleft:nside+nright,j)
      ; Include Gaussian weighting by boosting errors
      weight_yerr = local_yerr / sqrt(local_ker)
      ; Polynomial fit
      local_pfit  = poly_fit(local_x,local_y,n,meas=weight_yerr,$
                             yfit=local_yfit,chisq=local_chisq)
      ;local_red_chisq = local_chisq / total(local_ker)
      temp_yinterp = local_yfit(nleft)
      temp_chisq   = local_chisq / total(local_ker)
      ; If chisq is too large, try the next smaller Gaussian width
      if (temp_chisq gt max_chisq and j ne nsig-1) then begin
         j = j+1
      endif else begin
         fit_done = 1
      endelse
   endwhile
   ; Save the value of the polynomial fit at the current position
   yinterp(i) = temp_yinterp
   chisq(i)   = temp_chisq
   sigused(i) = siglist(j)
   ; Continue to next pixel or interval
   i = round(i + istep)
endwhile

; If computing Gaussian interpolation only at intervals
; use a spline to interpolate the remaining values
if keyword_set(interval) then begin 
   sample_ls = istart + round(istep*lindgen(nstep))
   yinterp   = interpol(yinterp(sample_ls),x(sample_ls),x,/spline)
endif

; If performing interpolation at a single point only,
; return scalar value.  Otherwise, return the entire
; interpolated array.
if keyword_set(x0) then begin

   ; If wrapping was done, restore original arrays   
   if (n_edge ne 0) then begin
      x    = x_orig
      y    = y_orig
      yerr = yerr_orig
   endif

   chisq = chisq(i0)
   return,yinterp(i0)

endif else begin

   ; If wrapping was done, omit extra pixels at the ends
   if (n_edge ne 0) then begin
      ilo  = n_edge
      ihi  = npix-1-n_edge
      x    = x_orig
      y    = y_orig
      yerr = yerr_orig
      yinterp = yinterp(ilo:ihi)
      chisq   = chisq(ilo:ihi)
      sigused = sigused(ilo:ihi)
   endif

   return,yinterp

endelse

end
